@@ -23,6 +23,7 @@
 import argparse
 import sys
 import json
+import simplejson
 import traceback
 import types
 import socket
@@ -42,7 +43,7 @@ from keys import *
 from proxy import *
 from profile import *
 
-from spv import SPVClient
+from virtualchain import SPVClient
 
 import pybitcoin
 import bitcoin
@@ -59,6 +60,7 @@ log = get_logger()
 
 import virtualchain
 
+'''
 def get_create_diff(blockchain_record):
     """
     Given a blockchain record, find its earliest history diff and its creation block number.
@@ -77,8 +79,9 @@ def get_create_diff(blockchain_record):
 
     create_diff = history[str(create_block_number)][0]
     return (create_block_number, create_diff)
+'''
 
-
+'''
 def get_reveal_txid(blockchain_record):
     """
     Given a blockchain record, find the transaction ID that revealed it
@@ -98,8 +101,9 @@ def get_reveal_txid(blockchain_record):
                     return reveal_block_number, str(history[block_number_str][i]['txid'])
 
     return (None, None)
+'''
 
-
+'''
 def get_serial_number(blockchain_record):
     """
     Calculate the serial number from a name's blockchain record.
@@ -133,28 +137,7 @@ def get_serial_number(blockchain_record):
         serial_number = str(create_block_number) + '-' + str(create_tx_index)
 
     return serial_number
-
-
-def get_block_from_consensus(consensus_hash, proxy=None):
-    """
-    Get a block ID from a consensus hash
-    """
-    if proxy is None:
-        proxy = get_default_proxy()
-
-    resp = proxy.get_block_from_consensus(consensus_hash)
-    if resp is None:
-        resp = {'error': 'No such consensus hash'}
-
-    elif type(resp) == list:
-        # backwards-compatibility
-        if len(resp) == 0:
-            resp = {'error': 'No data returned'}
-        else:
-            resp = resp[0]
-
-    return resp
-
+'''
 
 def txid_to_block_data(txid, bitcoind_proxy, proxy=None):
     """
@@ -194,7 +177,7 @@ def txid_to_block_data(txid, bitcoind_proxy, proxy=None):
     block_id = SPVClient.block_header_index(proxy.spv_headers_path, (untrusted_block_header_hex + "00").decode('hex'))
     if block_id < 0:
         # bad header
-        log.error("Block header '%s' is not in the SPV headers" % untrusted_block_header_hex)
+        log.error("Block header '%s' is not in the SPV headers (%s)" % (untrusted_block_header_hex, proxy.spv_headers_path))
         return (None, None, None)
 
     # block header is trusted.  Is the transaction data consistent with it?
@@ -215,7 +198,7 @@ def txid_to_block_data(txid, bitcoind_proxy, proxy=None):
     return (block_hash, block_data, tx_data)
 
 
-
+'''
 def txid_to_serial_number(txid, bitcoind_proxy, proxy=None):
     """
     Given a transaction ID, convert it into a serial number
@@ -244,8 +227,9 @@ def txid_to_serial_number(txid, bitcoind_proxy, proxy=None):
         # not actually present
         log.error("Transaction %s is not present in block %s (%s)" % (txid, block_id, block_hash))
 
+    # TODO: block_id?
     return "%s-%s" % (block_id, tx_index)
-
+'''
 
 
 def serial_number_to_tx(serial_number, bitcoind_proxy, proxy=None):
@@ -336,9 +320,8 @@ def parse_tx_op_return(tx):
             break
 
     if op_return is None:
-        pp = pprint.PrettyPrinter()
-        pp.pprint(tx)
         log.error("transaction has no OP_RETURN output")
+        log.debug("transaction has no OP_RETURN output:\n%s" % simplejson.dumps(tx))
         return (None, None)
 
     # [0] is OP_RETURN, [1] is the length; [2:4] are 'id', [4] is opcode
@@ -379,6 +362,7 @@ def get_consensus_hash_from_tx(tx):
         return consensus_hash
 
 
+'''
 def verify_consensus_hash_from_tx(tx, fqname, candidate_consensus_hash):
     """
     Given the SPV-verified transaction that encodes a consensus hash-bearing OP_RETURN,
@@ -421,8 +405,9 @@ def verify_consensus_hash_from_tx(tx, fqname, candidate_consensus_hash):
 
         log.error("NAME_UPDATE consensus hash mix mismatch: expected %s from %s, got %s" % (name_consensus_hash_mix.encode('hex'), candidate_consensus_hash, tx_name_consensus_hash_mix.encode('hex')))
         return None
+'''
 
-
+'''
 def get_name_creation_consensus_info(name, blockchain_record, bitcoind_proxy, proxy=None, serial_number=None):
     """
     Given the result of a call to get_name_blockchain_record,
@@ -501,8 +486,9 @@ def get_name_creation_consensus_info(name, blockchain_record, bitcoind_proxy, pr
         }
 
     return creation_info
+'''
 
-
+'''
 def get_name_reveal_consensus_info(name, blockchain_record, bitcoind_proxy, proxy=None):
     """
     Given a name, its blockchain record, and a bitcoind proxy, get information
@@ -557,8 +543,9 @@ def get_name_reveal_consensus_info(name, blockchain_record, bitcoind_proxy, prox
     }
 
     return ret
+'''
 
-
+'''
 def find_last_historical_op(history, opcode):
     """
     Given the blockchain history of a name, and the desired opcode name,
@@ -588,8 +575,9 @@ def find_last_historical_op(history, opcode):
                 return (int(prev_blocks[i + 1]), prev_op)
 
     return (None, None)
+'''
 
-
+'''
 def get_name_update_consensus_info(name, blockchain_record, bitcoind_proxy, proxy=None):
     """
     Given the result of a call to get_name_blockchain_record (an untrusted database record for a name),
@@ -666,7 +654,7 @@ def get_name_update_consensus_info(name, blockchain_record, bitcoind_proxy, prox
         }
 
         return update_info
-
+'''
 
 def snv_get_nameops_at(current_block_id, current_consensus_hash, block_id, consensus_hash, proxy=None):
     """
@@ -710,26 +698,8 @@ def snv_get_nameops_at(current_block_id, current_consensus_hash, block_id, conse
         else:
             nameops_hash = prev_nameops_hashes[ next_block_id ]
 
-        # print "prev_nameops_hashes[%s] = %s" % (next_block_id, nameops_hash)
-        """
-        ch_block_ids = []
-        while next_block_id - (2**(i+1) - 1) >= FIRST_BLOCK_MAINNET:
+        log.debug("nameops hash at %s: %s" % (next_block_id, nameops_hash))
 
-            i += 1
-            ch_block_ids.append(next_block_id - (2**i - 1))
-
-            if not prev_consensus_hashes.has_key(next_block_id - (2**i - 1)):
-                ch = str(get_consensus_at(next_block_id - (2**i - 1), proxy=proxy))
-
-                if ch != "None":
-                    prev_consensus_hashes[ next_block_id - (2**i - 1) ] = ch
-                    # print "prev_consensus_hashes[%s] = %s" % (next_block_id - (2**i - 1), ch)
-
-                else:
-                    # skip this one
-                    ch_block_ids.pop()
-                    break
-        """
         # find out which consensus hashes we'll need
         to_fetch = []
         ch_block_ids = []
@@ -747,13 +717,16 @@ def snv_get_nameops_at(current_block_id, current_consensus_hash, block_id, conse
         chs = {}
         if len(to_fetch) > 0:
             chs = get_consensus_hashes( to_fetch, proxy=proxy )
+            if 'error' in chs:
+                log.error("Failed to get consensus hashes for %s: %s" % (to_fetch, chs['error']))
+                return {'error': 'Failed to get consensus hashes'}
 
         prev_consensus_block_ids = []
         for b in ch_block_ids:
 
             # NOTE: we process to_fetch *in decreasing order* so we know when we're missing data
             if not chs.has_key(b) and not prev_consensus_hashes.has_key(b):
-                log.error("Missing consensus hash response for %s" % b)
+                log.error("Missing consensus hash response for %s (chs=%s, prev_chs=%s)" % (b, chs, prev_consensus_hashes))
                 return {'error': "Server did not reply valid data"}
 
             prev_consensus_block_ids.append(b)
@@ -777,7 +750,7 @@ def snv_get_nameops_at(current_block_id, current_consensus_hash, block_id, conse
         ch = virtualchain.StateEngine.make_snapshot_from_ops_hash(nameops_hash, prev_consensus_hashes_list)
         expected_ch = prev_consensus_hashes[ next_block_id ]
         if ch != expected_ch:
-            log.error("Consensus hash mismatch at %s: expected %s, got %s" % (next_block_id, expected_ch, ch))
+            log.error("Consensus hash mismatch at %s: expected %s, got %s (from %s, %s)" % (next_block_id, expected_ch, ch, nameops_hash, prev_consensus_hashes))
             return {'error': 'Consensus hash mismatch'}
 
         # advance!
@@ -800,6 +773,7 @@ def snv_get_nameops_at(current_block_id, current_consensus_hash, block_id, conse
     # get the final nameops
     historic_nameops = get_nameops_at(block_id, proxy=proxy)
     if type(historic_nameops) == dict and 'error' in historic_nameops:
+        log.error("Failed to get nameops at %s: %s" % (block_id, historic_nameops['error']))
         return {'error': 'BUG: no nameops found'}
 
     # sanity check...
@@ -812,7 +786,7 @@ def snv_get_nameops_at(current_block_id, current_consensus_hash, block_id, conse
             historic_op['op'] = NAME_OPCODES[ str(historic_op['opcode']) ]
 
     # check integrity
-    serialized_historic_nameops = [virtualchain.StateEngine.serialize_op(str(op['op'][0]), op, OPFIELDS, verbose=False) for op in historic_nameops]
+    serialized_historic_nameops = [virtualchain.StateEngine.serialize_op(str(op['op'][0]), op, OPFIELDS, verbose=True) for op in historic_nameops]
     historic_nameops_hash = virtualchain.StateEngine.make_ops_snapshot(serialized_historic_nameops)
 
     if not prev_nameops_hashes.has_key(block_id):
@@ -821,16 +795,19 @@ def snv_get_nameops_at(current_block_id, current_consensus_hash, block_id, conse
     if historic_nameops_hash != prev_nameops_hashes[ block_id ]:
         return {'error': 'Hash mismatch: name is not consistent with consensus hash'}
 
+    log.debug("%s nameops at %s" % (len(historic_nameops), block_id))
     return historic_nameops
 
 
-def snv_name_verify(name, current_block_id, current_consensus_hash, block_id, consensus_hash, proxy=None):
+def snv_name_verify(name, current_block_id, current_consensus_hash, block_id, consensus_hash, trusted_txid=None, trusted_txindex=None, proxy=None):
     """
     Use SNV to verify that a name existed at a particular block ID in the past,
     given a later known-good block ID and consensus hash (as well as the previous
     untrusted consensus hash)
 
-    Return the name's historic nameop on success
+    Return the name's historic nameop(s) on success.
+    If there are multiple matches, multiple nameops will be returned.
+    The return value takes the form of {'status': True, 'nameops': [...]}
     Return a dict with {'error'} on error
     """
 
@@ -841,22 +818,38 @@ def snv_name_verify(name, current_block_id, current_consensus_hash, block_id, co
     if 'error' in historic_nameops:
         return historic_nameops
 
+    matching_nameops = []
+
     # find the one we asked for
     for nameop in historic_nameops:
+        # select on more-accurate filters first
+        if trusted_txindex is not None and nameop['vtxindex'] == trusted_txindex:
+            matching_nameops = [nameop]
+            break
+
+        if trusted_txid is not None and nameop['txid'] == trusted_txid:
+            matching_nameops = [nameop]
+            break
+
         if 'name' not in nameop:
             continue
 
         if str(nameop['name']) == str(name):
             # success!
-            return nameop
+            matching_nameops.append(nameop)
+            continue
 
-    # not found
-    log.error("Not found at block %s: '%s'" % (block_id, name))
-    return {'error': 'Name not found'}
+    if len(matching_nameops) == 0:
+        # not found
+        log.error("Not found at block %s: '%s'" % (block_id, name))
+        return {'error': 'Name not found'}
+    else:
+        return {'status': True, 'nameops': matching_nameops}
 
 
 
-def snv_lookup(verify_name, verify_block_id, trusted_serial_number_or_txid_or_consensus_hash, proxy=None):
+
+def snv_lookup(verify_name, verify_block_id, trusted_serial_number_or_txid_or_consensus_hash, proxy=None, trusted_txid=None):
     """
     High-level call to simple name verification:
     Given a trusted serial number, txid, or consensus_hash, use it as a trust root to verify that
@@ -884,6 +877,10 @@ def snv_lookup(verify_name, verify_block_id, trusted_serial_number_or_txid_or_co
     then the transitive closure of all prior consensus hashes will be assumed valid as well.
     This means that the only way to drive the valid consensus hash from a prior invalid
     consensus hash is to force a hash collision somewhere in the transitive closure, which is infeasible.
+
+    NOTE: @trusted_txid is needed for isolating multiple operations in the same name within a single block.
+
+    Return the list of nameops in the given verify_block_id that match.
     """
 
     if proxy is None:
@@ -893,7 +890,7 @@ def snv_lookup(verify_name, verify_block_id, trusted_serial_number_or_txid_or_co
 
     bitcoind_proxy = get_bitcoind_client( config_path=proxy.conf['path'] )
     trusted_serial_number = None
-    trusted_txid = None
+    trusted_tx_index = None
     trusted_consensus_hash = None
     trusted_block_id = None
 
@@ -962,9 +959,14 @@ def snv_lookup(verify_name, verify_block_id, trusted_serial_number_or_txid_or_co
 
     # go verify the name
     verify_consensus_hash = get_consensus_at(verify_block_id, proxy=proxy)
-    historic_namerec = snv_name_verify(verify_name, trusted_block_id, trusted_consensus_hash, verify_block_id, verify_consensus_hash)
+    historic_namerecs = snv_name_verify(verify_name, trusted_block_id, trusted_consensus_hash, verify_block_id, verify_consensus_hash, trusted_txid=trusted_txid, trusted_txindex=trusted_tx_index)
+
+    if 'error' in historic_namerecs:
+        return historic_namerecs
+
+    else:
+        return historic_namerecs['nameops']
 
-    return historic_namerec
 
 # backwards compatibility
 lookup_snv = snv_lookup
