@@ -20,9 +20,11 @@
     along with Blockstack-client.  If not, see <http://www.gnu.org/licenses/>.
 """
 
+import time
 import argparse
 import sys
 import json
+import simplejson
 import traceback
 import os
 import re
@@ -30,6 +32,8 @@ import errno
 import pybitcoin
 import subprocess
 import shutil
+import virtualchain
+
 from keylib import ECPrivateKey
 from socket import error as socket_error
 from time import sleep
@@ -54,24 +58,24 @@ from pybitcoin import make_send_to_address_tx
 from pybitcoin import BlockcypherClient
 from pybitcoin.rpc.bitcoind_client import BitcoindClient
 
-# Hack around absolute paths
-current_dir = os.path.abspath(os.path.dirname(__file__))
-parent_dir = os.path.abspath(current_dir + "/../")
-sys.path.insert(0, parent_dir)
-
 from .backend.crypto.utils import get_address_from_privkey, get_pubkey_from_privkey
 from .backend.crypto.utils import aes_encrypt, aes_decrypt
-from .backend.blockchain import get_balance, is_address_usable, get_tx_fee
+from .backend.blockchain import get_balance, is_address_usable, get_tx_fee, get_block_height
 from .utils import satoshis_to_btc, btc_to_satoshis, exit_with_error, print_result
 
+from .keys import *
+
 import config
-from .config import WALLET_PATH, WALLET_PASSWORD_LENGTH, CONFIG_PATH, CONFIG_DIR, CONFIG_FILENAME, WALLET_FILENAME, MINIMUM_BALANCE
+from .config import WALLET_PATH, WALLET_PASSWORD_LENGTH, CONFIG_PATH, CONFIG_DIR, CONFIG_FILENAME, WALLET_FILENAME, MINIMUM_BALANCE, \
+        WALLET_DECRYPT_MAX_TRIES, WALLET_DECRYPT_BACKOFF_RESET
 
 from .proxy import get_names_owned_by_address, get_default_proxy, get_name_cost
 from .rpc import local_rpc_connect, start_rpc_endpoint
 
 log = config.get_logger()
 
+DECRYPT_ATTEMPTS = 0
+LAST_DECRYPT_ATTEMPT = 0
 
 class HDWallet(object):
 
@@ -179,9 +183,15 @@ class HDWallet(object):
         return None
 
 
-def make_wallet( password, hex_privkey=None, payment_privkey=None, owner_privkey=None, data_privkey=None, config_path=CONFIG_PATH ):
+def make_wallet( password, hex_privkey=None, payment_privkey_info=None, owner_privkey_info=None, data_privkey_info=None, config_path=CONFIG_PATH ):
     """
-    Make a wallet structure
+    Make a wallet structure.
+    By default, the owner and payment keys will be key bundles set up to require 2-of-3 signatures.
+    @payment_privkey_info, @owner_privkey_info, and @data_privkey_info can either be individual private keys, or 
+    dicts with {'redeem_script': ..., 'private_keys': ...} defined.
+
+    Return the new wallet on success.
+    Return {'error': ...} on failure
     """
 
     hex_password = hexlify(password)
@@ -196,87 +206,229 @@ def make_wallet( password, hex_privkey=None, payment_privkey=None, owner_privkey
     encrypted_key = aes_encrypt(hex_privkey, hex_password)
     data['encrypted_master_private_key'] = encrypted_key
 
-    if payment_privkey is None:
-        data['payment_addresses'] = [child[0][0]]
-    else:
-        try:
-            data['payment_addresses'] = [pybitcoin.BitcoinPrivateKey(payment_privkey).public_key().address()]
-        except:
-            return {'error': 'Invalid payment private key'}
+    multisig = False
+    curr_height = get_block_height( config_path=config_path )
+    if curr_height >= config.EPOCH_HEIGHT_MINIMUM:
+        # safe to use multisig
+        multisig = True
 
-        data['encrypted_payment_privkey'] = aes_encrypt(payment_privkey, hex_password)
+    # default to 2-of-3 multisig key info if data isn't given
+    if payment_privkey_info is None:
+        if multisig:
+            payment_privkey_info = virtualchain.make_multisig_wallet( 2, 3 )
+        else:
+            payment_privkey_info = virtualchain.BitcoinPrivateKey().to_wif()
 
-    if owner_privkey is None:
-        data['owner_addresses'] = [child[1][0]]
-    else:
-        try:
-            data['owner_addresses'] = [pybitcoin.BitcoinPrivateKey(owner_privkey).public_key().address()]
-        except:
-            return {'error': 'Invalid payment private key'}
+    if not is_singlesig(payment_privkey_info) and not is_multisig(payment_privkey_info):
+        return {'error': 'Payment private key info must be either a single private key or a multisig bundle'}
 
-        data['encrypted_owner_privkey'] = aes_encrypt(owner_privkey, hex_password)
+    if not multisig and is_multisig(payment_privkey_info):
+        return {'error': 'Multisig payment private key info is not supported'}
 
-    data_keypair = child[2]
-    if data_privkey is None:
-        data['data_pubkeys'] = [ECPrivateKey(data_keypair[1]).public_key().to_hex()]
-    else:
-        try:
-            data['data_pubkeys'] = [ECPrivateKey(data_privkey).public_key().to_hex()]
-        except:
-            return {'error': 'Invalid data private key'}
+    if owner_privkey_info is None:
+        if multisig:
+            owner_privkey_info = virtualchain.make_multisig_wallet( 2, 3 )
+        else:
+            owner_privkey_info = virtualchain.BitcoinPrivateKey().to_wif()
 
-        data['encrypted_data_privkey'] = aes_encrypt(data_privkey, hex_password)
+    if not is_singlesig(owner_privkey_info) and not is_multisig(owner_privkey_info):
+        return {'error': 'Owner private key info must be either a single private key or a multisig bundle'}
 
+    if not multisig and is_multisig(owner_privkey_info):
+        return {'error': 'Multisig owner private key info is not supported'}
+
+    if data_privkey_info is None:
+        # TODO: for now, this must be a single private key 
+        data_privkey_info = child[2][1]
+
+    elif not is_singlesig(data_privkey_info):
+        return {'error': 'Data private key info must be a single private key'}
+
+    enc_payment_info = encrypt_private_key_info( payment_privkey_info, password )
+    if 'error' in enc_payment_info:
+        return {'error': enc_payment_info['error']}
+
+    enc_owner_info = encrypt_private_key_info( owner_privkey_info, password )
+    if 'error' in enc_owner_info:
+        return {'error': enc_owner_info['error']}
+
+    enc_data_info = encrypt_private_key_info( data_privkey_info, password )
+    if 'error' in enc_data_info:
+        return {'error': enc_data_info['error']}
+
+    payment_addr = enc_payment_info['encrypted_private_key_info']['address']
+    owner_addr = enc_owner_info['encrypted_private_key_info']['address']
+
+    enc_payment_info = enc_payment_info['encrypted_private_key_info']['private_key_info']
+    enc_owner_info = enc_owner_info['encrypted_private_key_info']['private_key_info']
+    enc_data_info = enc_data_info['encrypted_private_key_info']['private_key_info']
+
+    data['encrypted_payment_privkey'] = enc_payment_info
+    data['payment_addresses'] = [payment_addr]
+
+    data['encrypted_owner_privkey'] = enc_owner_info
+    data['owner_addresses'] = [owner_addr]
+
+    data['encrypted_data_privkey'] = enc_data_info
+    data['data_pubkeys'] = [virtualchain.BitcoinPrivateKey(data_privkey_info).public_key().to_hex()]
     data['data_pubkey'] = data['data_pubkeys'][0]
-
     return data
 
 
-def decrypt_wallet( data, password, config_path=CONFIG_PATH ):
+def log_failed_decrypt(max_tries=WALLET_DECRYPT_MAX_TRIES):
     """
-    Decrypt a wallet's encrypted fields
+    Record that we tried (and failed)
+    to decrypt a wallet.  Determine
+    how long we should wait before 
+    allowing another attempt.
+
+    If we tried many times, then use
+    exponential backoff to limit brute-forces
+
+    Return the interval of time to sleep
+    """
+    global DECRYPT_ATTEMPTS
+    global LAST_DECRYPT_ATTEMPT
+    global NEXT_DECRYPT_ATTEMPT
+
+    if LAST_DECRYPT_ATTEMPT + WALLET_DECRYPT_BACKOFF_RESET < time.time():
+        # haven't tried in a while
+        DECRYPT_ATTEMPTS = 0
+        NEXT_DECRYPT_ATTEMPT = 0
+        return 
+
+    DECRYPT_ATTEMPTS += 1
+    LAST_DECRYPT_ATTEMPT = time.time()
+    
+    if DECRYPT_ATTEMPTS > max_tries:
+
+        interval = 2**(DECRYPT_ATTEMPTS - max_tries + 1)
+        NEXT_DECRYPT_ATTEMPT = time.time() + interval
+   
+    return
+
+
+def can_attempt_decrypt( max_tries=WALLET_DECRYPT_MAX_TRIES ):
+    """
+    Can we attempt a decryption?
+    Has enough time passed since the last guess?
+    """
+    global DECRYPT_ATTEMPTS
+    global LAST_DECRYPT_ATTEMPT
+    global NEXT_DECRYPT_ATTEMPT
+
+    if LAST_DECRYPT_ATTEMPT + WALLET_DECRYPT_BACKOFF_RESET < time.time():
+        # haven't tried in a while
+        DECRYPT_ATTEMPTS = 0
+        NEXT_DECRYPT_ATTEMPT = 0
+        return True
+
+    if NEXT_DECRYPT_ATTEMPT < time.time():
+        return True
+
+    else:
+        return False
+
+
+def time_until_next_decrypt_attempt():
+    """
+    When can we try to decrypt next?
+    """
+    global NEXT_DECRYPT_ATTEMPT
+    if NEXT_DECRYPT_ATTEMPT == 0:
+        return 0
+
+    if NEXT_DECRYPT_ATTEMPT < time.time():
+        return 0
+
+    return NEXT_DECRYPT_ATTEMPT - time.time()
+
+
+def decrypt_wallet( data, password, config_path=CONFIG_PATH, max_tries=WALLET_DECRYPT_MAX_TRIES ):
+    """
+    Decrypt a wallet's encrypted fields.
+
+    After WALLET_DECRYPT_MAX_TRIES failed attempts, start doing exponential backoff
+    to prevent brute-force attacks.
+
     Return a dict with the decrypted fields on success
     Return {'error': ...} on failure
     """
     hex_password = hexlify(password)
     wallet = None
 
+    if not can_attempt_decrypt( max_tries=max_tries ):
+        return {'error': 'Cannot decrypt at this time.  Try again in %s seconds' % time_until_next_decrypt_attempt()}
+
     try:
         hex_privkey = aes_decrypt(data['encrypted_master_private_key'], hex_password)
         wallet = HDWallet(hex_privkey)
     except Exception, e:
         if os.environ.get("BLOCKSTACK_DEBUG", None) == "1":
             log.exception(e)
-        return {'error': 'Incorrect password'}
-    
+
+        ret = {'error': 'Incorrect password'}
+        log_failed_decrypt( max_tries=max_tries )
+        if not can_attempt_decrypt( max_tries=max_tries ):
+            log.debug("Incorrect password; using exponential backoff")
+            ret['error'] = 'Incorrect password.  Try again in %s seconds' % time_until_next_decrypt_attempt()
+
+        return ret
+   
+    # legacy compat: use the master private key to generate child keys.
+    # If the specific key they are purposed for is not defined in the wallet,
+    # then they are used in its place.
+    # This is because originally, the master private key was used to derive
+    # the owner, payment, and data private keys; not all wallets define
+    # these keys separately (and have instead relied on us being able to
+    # generate them from the master private key).
     child = wallet.get_child_keypairs(count=3, include_privkey=True)
     payment_keypair = child[0]
     owner_keypair = child[1]
     data_keypair = child[2]
     data_pubkey = ECPrivateKey( data_keypair[1] ).public_key().to_hex()
 
+    multisig = False
+    curr_height = get_block_height( config_path=config_path )
+    if curr_height >= config.EPOCH_HEIGHT_MINIMUM:
+        # safe to use multisig 
+        multisig = True
+
     ret = {}
-    keynames = ['payment_privkey', 'owner_privkey', 'data_privkey']
+    keynames = ['payment', 'owner', 'data']
     for i in xrange(0, len(keynames)):
 
         keyname = keynames[i]
+        keyname_privkey = "%s_privkey" % keyname
+        keyname_addresses = "%s_addresses" % keyname
+
         child_keypair = child[i]
-        encrypted_keyname = "encrypted_%s" % keyname
+        encrypted_keyname = "encrypted_%s_privkey" % keyname
 
         if data.has_key(encrypted_keyname):
-            try:
-                privkey = aes_decrypt(data[encrypted_keyname], hex_password)
-            except Exception, e:
-                log.exception(e)
-                return {'error': 'Incorrect password'}
+            # This key was explicitly defined in the wallet.
+            # It is not guaranteed to be a child key of the
+            # master private key.
+            field = decrypt_private_key_info( data[encrypted_keyname], password )
+            if 'error' in field:
+                log.debug("Failed to decrypt '%s': %s" % (encrypted_keyname, field['error']))
+                return field
+
+            ret[keyname_privkey] = field['private_key_info']
+            ret[keyname_addresses] = [field['address']]
 
-            ret[keyname] = privkey
         else:
-            ret[keyname] = child_keypair[1]
+            # Legacy: this key is not defined in the wallet.
+            # Derive it from the master key.
+            ret[keyname_privkey] = data_keypair[1]
+            ret[keyname_addresses] = [virtualchain.BitcoinPrivateKey(ret[keyname_privkey]).public_key().address()]
+
+        # this can't be multisig if it's not yet supported 
+        if not is_singlesig( ret[keyname_privkey] ) and not multisig:
+            log.error("Invalid wallet data for '%s'" % keyname_privkey)
+            return {'error': 'Invalid wallet'}
 
     ret['hex_privkey'] = hex_privkey
-    ret['payment_addresses'] = [pybitcoin.BitcoinPrivateKey(ret['payment_privkey']).public_key().address()]
-    ret['owner_addresses'] = [pybitcoin.BitcoinPrivateKey(ret['owner_privkey']).public_key().address()]
     ret['data_pubkeys'] = [ECPrivateKey(ret['data_privkey']).public_key().to_hex()]
     ret['data_pubkey'] = ret['data_pubkeys'][0]
 
@@ -298,7 +450,7 @@ def write_wallet( data, path=None, config_dir=CONFIG_DIR ):
     return True
 
 
-def make_wallet_password( password=None ):
+def make_wallet_password( prompt=None, password=None ):
     """
     Make a wallet password:
     prompt for a wallet, and ensure it's the right length.
@@ -313,6 +465,9 @@ def make_wallet_password( password=None ):
         return {'status': True, 'password': password}
 
     else:
+        if prompt:
+            print prompt
+
         p1 = getpass("Enter new password: ")
         p2 = getpass("Confirm new password: ")
         if p1 != p2:
@@ -325,10 +480,11 @@ def make_wallet_password( password=None ):
             return {'status': True, 'password': p1}
 
 
-def initialize_wallet( password="", interactive=True, hex_privkey=None, config_dir=CONFIG_DIR, wallet_path=None ):
+def initialize_wallet( password="", interactive=True, hex_privkey=None, config_dir=CONFIG_DIR, wallet_path=None, owner_privkey_info=None, payment_privkey_info=None, data_privkey_info=None ):
     """
-    Initialize the wallet,
+    Initialize a wallet,
     interatively if need be.
+    Save it to @wallet_path
     Return a dict with the wallet password and master private key.
     Return {'error': ...} on error
     """
@@ -336,16 +492,18 @@ def initialize_wallet( password="", interactive=True, hex_privkey=None, config_d
         wallet_path = os.path.join(config_dir, WALLET_FILENAME)
 
     config_path = os.path.join(config_dir, CONFIG_FILENAME)
-        
+    
     if not interactive and (password is None or len(password) == 0):
         raise Exception("Non-interactive wallet initialization requires a password of length %s or greater" % WALLET_PASSWORD_LENGTH)
 
     result = {}
-    print "Initializing new wallet ..."
+
+    if interactive:
+        print "Initializing new wallet ..."
 
     try:
         if interactive:
-            while len(password) < WALLET_PASSWORD_LENGTH:
+            while password is None or len(password) < WALLET_PASSWORD_LENGTH:
                 res = make_wallet_password(password)
                 if 'error' in res:
                     print res['error']
@@ -359,16 +517,20 @@ def initialize_wallet( password="", interactive=True, hex_privkey=None, config_d
             temp_wallet = HDWallet()
             hex_privkey = temp_wallet.get_master_privkey()
 
-        wallet = make_wallet( password, hex_privkey=hex_privkey, config_path=config_path )
-        write_wallet( wallet, path=wallet_path ) 
+        wallet = make_wallet( password, hex_privkey=hex_privkey, config_path=config_path, owner_privkey_info=owner_privkey_info, payment_privkey_info=payment_privkey_info, data_privkey_info=data_privkey_info )
+        if 'error' in wallet:
+            log.error("make_wallet failed: %s" % wallet['error'])
+            return wallet
 
-        print "Wallet created. Make sure to backup the following:"
+        write_wallet( wallet, path=wallet_path ) 
 
         result['wallet_password'] = password
         result['master_private_key'] = hex_privkey
-        print_result(result)
 
         if interactive:
+            print "Wallet created. Make sure to backup the following:"
+            print_result(result)
+
             input_prompt = "Have you backed up the above private key? (y/n): "
             user_input = raw_input(input_prompt)
             user_input = user_input.lower()
@@ -396,7 +558,7 @@ def wallet_exists(config_dir=CONFIG_DIR, wallet_path=None):
 
 def load_wallet( password=None, config_dir=CONFIG_DIR, wallet_path=None, include_private=False ):
     """
-    Get the wallet from disk, and unlock it.
+    Get a wallet from disk, and unlock it.
     Return {'status': True, 'wallet': ...} on success
     Return {'error': ...} on error
     """
@@ -408,10 +570,9 @@ def load_wallet( password=None, config_dir=CONFIG_DIR, wallet_path=None, include
     if password is None:
         password = getpass("Enter wallet password: ")
 
-    file = open(wallet_path, 'r')
-    data = file.read()
-    data = json.loads(data)
-    file.close()
+    with open(wallet_path, 'r') as f:
+        data = f.read()
+        data = json.loads(data)
 
     wallet = decrypt_wallet( data, password, config_path=config_path )
     if 'error' in wallet:
@@ -421,23 +582,23 @@ def load_wallet( password=None, config_dir=CONFIG_DIR, wallet_path=None, include
         return {'status': True, 'wallet': wallet}
     
 
-def unlock_wallet(display_enabled=False, password=None, config_dir=CONFIG_DIR, wallet_path=None ):
+def unlock_wallet( password=None, config_dir=CONFIG_DIR, wallet_path=None ):
     """
     Unlock the wallet.
     Save the wallet to the RPC daemon on success.
-    exit on error (e.g. incorrect password)
-    Return {'status': True} on success
+
+    If this wallet is in legacy format, then it will
+    be migrated to the latest format and the legacy
+    copy backed up.
+
+    Return {'status': True, 'addresses': ...} on success
     return {'error': ...} on error
     """
     config_path = os.path.join( config_dir, CONFIG_FILENAME )
     if wallet_path is None:
         wallet_path = os.path.join( config_dir, WALLET_FILENAME )
 
-    if walletUnlocked(config_dir):
-        if display_enabled:
-            payment_address, owner_address, data_pubkey = get_addresses_from_file(wallet_path=wallet_path)
-            display_wallet_info(payment_address, owner_address, data_pubkey, config_path=config_path)
-
+    if is_wallet_unlocked(config_dir):
         return {'status': True}
 
     else:
@@ -451,18 +612,15 @@ def unlock_wallet(display_enabled=False, password=None, config_dir=CONFIG_DIR, w
                 data = json.loads(data)
 
             wallet = decrypt_wallet( data, password, config_path=config_path )
-            if 'error' in wallet.keys():
-                log.error("Incorrect password")
-                return {'error': 'Incorrect password'}
-
-            if display_enabled:
-                display_wallet_info( wallet['payment_addresses'][0], wallet['owner_addresses'][0], wallet['data_pubkeys'][0], config_path=config_path )
+            if 'error' in wallet:
+                log.error("Failed to decrypt wallet: %s" % wallet['error'])
+                return wallet
 
             # may need to migrate data_pubkey into wallet.json
             _, _, onfile_data_pubkey = get_addresses_from_file(wallet_path=wallet_path)
             if onfile_data_pubkey is None:
 
-                # make a data keypair 
+                # make a data keypair (always the third child (index 2) of the HDWallet) 
                 w = HDWallet(wallet['hex_privkey'])
                 child = w.get_child_keypairs(count=3, include_privkey=True)
                 data_keypair = child[2]
@@ -472,8 +630,8 @@ def unlock_wallet(display_enabled=False, password=None, config_dir=CONFIG_DIR, w
                 wallet['data_pubkey'] = wallet['data_pubkeys'][0]
 
                 # set addresses 
-                wallet['payment_addresses'] = [pybitcoin.BitcoinPrivateKey(wallet['payment_privkey']).public_key().address()]
-                wallet['owner_addresses'] = [pybitcoin.BitcoinPrivateKey(wallet['owner_privkey']).public_key().address()]
+                wallet['payment_addresses'] = [get_privkey_info_address( wallet['payment_privkey'] )]
+                wallet['owner_addresses'] = [get_privkey_info_address( wallet['owner_privkey'] )]
 
                 # save!
                 encrypted_wallet = make_wallet( password, hex_privkey=wallet['hex_privkey'],
@@ -482,6 +640,10 @@ def unlock_wallet(display_enabled=False, password=None, config_dir=CONFIG_DIR, w
                                                           data_privkey=wallet['data_privkey'],
                                                           config_path=config_path )
 
+                if 'error' in encrypted_wallet:
+                    log.error("Failed to make wallet: %s" % encrypted_wallet['error'])
+                    return encrypted_wallet
+
                 write_wallet( encrypted_wallet, path=wallet_path + ".tmp" )
                 legacy_path = wallet_path + ".legacy"
                 if os.path.exists(wallet_path):
@@ -491,6 +653,7 @@ def unlock_wallet(display_enabled=False, password=None, config_dir=CONFIG_DIR, w
                         i = 1
                         while os.path.exists(legacy_path):
                             legacy_path = wallet_path + ".legacy.%s" % i
+                            i += 1
 
                         shutil.move( wallet_path, legacy_path )
 
@@ -498,21 +661,33 @@ def unlock_wallet(display_enabled=False, password=None, config_dir=CONFIG_DIR, w
                 log.debug("Migrated wallet %s (legacy wallet backed up to %s)" % (wallet_path, legacy_path))
 
             # save!
-            res = save_keys_to_memory( [wallet['payment_addresses'][0], wallet['payment_privkey']],
-                                       [wallet['owner_addresses'][0], wallet['owner_privkey']],
-                                       [wallet['data_pubkeys'][0], wallet['data_privkey']],
-                                       config_dir=config_dir )
+            try:
+                res = save_keys_to_memory( [wallet['payment_addresses'][0], wallet['payment_privkey']],
+                                           [wallet['owner_addresses'][0], wallet['owner_privkey']],
+                                           [wallet['data_pubkeys'][0], wallet['data_privkey']],
+                                           config_dir=config_dir )
+
+            except KeyError, ke:
+                if os.environ.get("BLOCKSACK_DEBUG", None) == "1":
+                    log.error("data: %s\n" % simplejson.dumps(wallet, indent=4, sort_keys=True))
+                raise
 
             if 'error' in res:
                 return res
 
-            return {'status': True}
+            addresses = {
+                "payment_address": wallet['payment_addresses'][0],
+                "owner_address": wallet['owner_addresses'][0],
+                "data_pubkey": wallet['data_pubkeys'][0]
+            }
+
+            return {'status': True, "addresses": addresses}
 
         except KeyboardInterrupt:
             return {'error': 'Interrupted'}
 
 
-def walletUnlocked(config_dir=CONFIG_DIR):
+def is_wallet_unlocked(config_dir=CONFIG_DIR):
     """
     Determine whether or not the wallet is unlocked.
     Do so by asking the local RPC backend daemon
@@ -630,6 +805,11 @@ def get_names_owned(address, proxy=None):
 def save_keys_to_memory(payment_keypair, owner_keypair, data_keypair, config_dir=CONFIG_DIR):
     """
     Save keys to the running RPC backend
+    Each keypair must be a list or tuple with 2 items: the address, and the private key information.
+    (Note that the private key information can be a multisig info dict).
+
+    Return {'status': True} on success
+    Return {'error': ...} on error
     """
     proxy = local_rpc_connect(config_dir=config_dir)
 
@@ -646,20 +826,30 @@ def get_addresses_from_file(config_dir=CONFIG_DIR, wallet_path=None):
     """
     Load up the set of addresses from the wallet
     Not all fields may be set in older wallets.
-    """
-    if wallet_path is None:
-        wallet_path = os.path.join(config_dir, WALLET_FILENAME)
+    """ 
 
-    file = open(wallet_path, 'r')
-    data = file.read()
-    data = json.loads(data)
-    file.close()
-    
     data_pubkey = None
     payment_address = None
     owner_address = None
 
-    # extract addresses 
+    if wallet_path is None:
+        wallet_path = os.path.join(config_dir, WALLET_FILENAME)
+
+    if not os.path.exists(wallet_path):
+        log.error("No such file or directory: %s" % wallet_path)
+        return None, None, None
+
+    with open(wallet_path, 'r') as f:
+        data = f.read()
+
+    try:
+        data = json.loads(data)
+    except:
+        log.error("Invalid wallet data: not JSON (in %s)" % wallet_path) 
+        return None, None, None 
+   
+    # extract addresses
+    # TODO: schema
     if data.has_key('payment_addresses'):
         payment_address = data['payment_addresses'][0]
     if data.has_key('owner_addresses'):
@@ -741,7 +931,7 @@ def get_total_balance(config_path=CONFIG_PATH, wallet_path=WALLET_PATH):
     return total_balance, payment_addresses
 
 
-def dump_wallet(config_path=CONFIG_PATH, password=None):
+def dump_wallet(config_path=CONFIG_PATH, wallet_path=None, password=None):
     """
     Load the wallet private keys.
     Return {'status': True, 'wallet': wallet} on success
@@ -750,13 +940,14 @@ def dump_wallet(config_path=CONFIG_PATH, password=None):
     config_dir = os.path.dirname(config_path)
     start_rpc_endpoint(config_dir)
 
-    wallet_path = os.path.join(config_dir, WALLET_FILENAME)
-    if not os.path.exists(wallet_path):
-        res = initialize_wallet(wallet_path=wallet_path, password=password)
-        if 'error' in res:
-            return res
+    if wallet_path is None:
+        wallet_path = os.path.join(config_dir, WALLET_FILENAME)
+        if not os.path.exists(wallet_path):
+            res = initialize_wallet(wallet_path=wallet_path, password=password)
+            if 'error' in res:
+                return res
 
-    if not walletUnlocked(config_dir=config_dir):
+    if not is_wallet_unlocked(config_dir=config_dir):
         res = unlock_wallet(config_dir=config_dir, password=password)
         if 'error' in res:
             return res
